#include <gtk/gtk.h>
#include <gmodule.h>

/* MOTIF COLOR CALCULATION CONSTANTS */
#define XmCOLOR_LITE_SEL_FACTOR 15
#define XmCOLOR_LITE_BS_FACTOR 40
#define XmCOLOR_LITE_TS_FACTOR 20
#define XmCOLOR_LO_SEL_FACTOR 15
#define XmCOLOR_LO_BS_FACTOR 60
#define XmCOLOR_LO_TS_FACTOR 50
#define XmCOLOR_HI_SEL_FACTOR 15
#define XmCOLOR_HI_BS_FACTOR 40
#define XmCOLOR_HI_TS_FACTOR 60
#define XmCOLOR_DARK_SEL_FACTOR 15
#define XmCOLOR_DARK_BS_FACTOR 30
#define XmCOLOR_DARK_TS_FACTOR 50
#define XmRED_LUMINOSITY 0.30
#define XmGREEN_LUMINOSITY 0.59
#define XmBLUE_LUMINOSITY 0.11
#define XmINTENSITY_FACTOR 75
#define XmLIGHT_FACTOR 0
#define XmLUMINOSITY_FACTOR 25
#define XmMAX_SHORT 65535
#define XmDEFAULT_DARK_THRESHOLD 20
#define XmDEFAULT_LIGHT_THRESHOLD 93
#define XmDEFAULT_FOREGROUND_THRESHOLD 70
#define XmCOLOR_PERCENTILE   (XmMAX_SHORT / 100)

#define XmCOLOR_LITE_THRESHOLD   XmDEFAULT_LIGHT_THRESHOLD* XmCOLOR_PERCENTILE
#define XmCOLOR_DARK_THRESHOLD   XmDEFAULT_DARK_THRESHOLD* XmCOLOR_PERCENTILE
#define XmFOREGROUND_THRESHOLD   XmDEFAULT_FOREGROUND_THRESHOLD* XmCOLOR_PERCENTILE

#define MOTIF_FG_COLOR 1
#define MOTIF_SEL_COLOR 2 
#define MOTIF_BS_COLOR 3
#define MOTIF_TS_COLOR 4

/* Theme functions to export */
void                theme_init(GtkThemeEngine * engine);
void                theme_exit(void);

/* Exported vtable from th_draw */

extern GtkStyleClass motif_default_class;

/* internals */

static guint
theme_parse_rc_style(GScanner * scanner,
		     GtkRcStyle * rc_style)
{
  guint               token;

  token = g_scanner_peek_next_token(scanner);
  while (token != G_TOKEN_RIGHT_CURLY)
    {
      switch (token)
	{
	default:
	  g_scanner_get_next_token(scanner);
	  token = G_TOKEN_RIGHT_CURLY;
	  break;
	}

      if (token != G_TOKEN_NONE)
	return token;

      token = g_scanner_peek_next_token(scanner);
    }

  g_scanner_get_next_token(scanner);

  rc_style->engine_data = NULL;

  return G_TOKEN_NONE;
}

static void
theme_merge_rc_style(GtkRcStyle * dest,
		     GtkRcStyle * src)
{
}

static void
theme_rc_style_to_style(GtkStyle * style,
			GtkRcStyle * rc_style)
{
  style->klass = &motif_default_class;
}

static void
theme_duplicate_style(GtkStyle * dest,
		      GtkStyle * src)
{
}

/* TODO: INSERT MOTIF CALCULATIONS CODE HERE */
static int 
Brightness(
     GdkColor *color )
{
	int brightness;
	int intensity;
	int light;
	int luminosity, maxprimary, minprimary;
	int red = color->red;
	int green = color->green;
	int blue = color->blue;

	intensity = (red + green + blue) / 3;

	/* 
	 * The casting nonsense below is to try to control the point at
	 * the truncation occurs.
	 */

	luminosity = (int) ((XmRED_LUMINOSITY * (float) red)
		+ (XmGREEN_LUMINOSITY * (float) green)
		+ (XmBLUE_LUMINOSITY * (float) blue));

	maxprimary = ( (red > green) ?
					( (red > blue) ? red : blue ) :
					( (green > blue) ? green : blue ) );

	minprimary = ( (red < green) ?
					( (red < blue) ? red : blue ) :
					( (green < blue) ? green : blue ) );

	light = (minprimary + maxprimary) / 2;

	brightness = ( (intensity * XmINTENSITY_FACTOR) +
				   (light * XmLIGHT_FACTOR) +
				   (luminosity * XmLUMINOSITY_FACTOR) ) / 100;
	return(brightness);
}


/* COLORTYPE IS: */
/* MOTIF_FG_COLOR  */
/* MOTIF_SEL_COLOR   */
/* MOTIF_BS_COLOR  */
/* MOTIF_TS_COLOR  */
static void 
CalculateColorsForMediumBackground
(GdkColor *dest, GdkColor *src, int colortype)
{
	int brightness = Brightness(src);
	int color_value, f;


    if (colortype==MOTIF_FG_COLOR) {
	if (brightness > XmFOREGROUND_THRESHOLD)
	{
		dest->red = 0;
		dest->green = 0;
		dest->blue = 0;
	}
	else
	{
		dest->red = XmMAX_SHORT;
		dest->green = XmMAX_SHORT;
		dest->blue = XmMAX_SHORT;
	}
     }

	if (colortype==MOTIF_SEL_COLOR)
	{
		f = XmCOLOR_LO_SEL_FACTOR + (brightness
			* ( XmCOLOR_HI_SEL_FACTOR - XmCOLOR_LO_SEL_FACTOR )
			/ XmMAX_SHORT );

		color_value = src->red;
		color_value -= (color_value * f) / 100;
		dest->red = color_value;

		color_value = src->green;
		color_value -= (color_value * f) / 100;
		dest->green = color_value;

		color_value = src->blue;
		color_value -= (color_value * f) / 100;
		dest->blue = color_value;
	}

	if (colortype==MOTIF_BS_COLOR)
	{
		f = XmCOLOR_LO_BS_FACTOR + (brightness 
			* ( XmCOLOR_HI_BS_FACTOR - XmCOLOR_LO_BS_FACTOR )
			/ XmMAX_SHORT);

		color_value = src->red;
		color_value -= (color_value * f) / 100;
		dest->red = color_value;

		color_value = src->green;
		color_value -= (color_value * f) / 100;
		dest->green = color_value;

		color_value = src->blue;
		color_value -= (color_value * f) / 100;
		dest->blue = color_value;
	}

	if (colortype==MOTIF_TS_COLOR)
	{
		f = XmCOLOR_LO_TS_FACTOR + (brightness
			* ( XmCOLOR_HI_TS_FACTOR - XmCOLOR_LO_TS_FACTOR )
			/ XmMAX_SHORT);

		color_value = src->red;
		color_value += f * ( XmMAX_SHORT - color_value ) / 100;
		dest->red = color_value;

		color_value = src->green;
		color_value += f * ( XmMAX_SHORT - color_value ) / 100;
		dest->green = color_value;

		color_value = src->blue;
		color_value += f * ( XmMAX_SHORT - color_value ) / 100;
		dest->blue = color_value;
	}
}




static void
calculate_colours(GtkStyle *style,
                      GtkStateType i)
{
  GdkGCValuesMask gc_values_mask = GDK_GC_FOREGROUND;
  GdkGCValues gc_values;

  /* Compute new colours for "light" and "dark".  */
  /* THIS IS MOTIF'S TS AND BS */
  CalculateColorsForMediumBackground(&style->light[i], &style->bg[i], MOTIF_TS_COLOR);
  CalculateColorsForMediumBackground(&style->dark[i], &style->bg[i], MOTIF_BS_COLOR);
  CalculateColorsForMediumBackground(&style->mid[i], &style->bg[i], MOTIF_SEL_COLOR);
  CalculateColorsForMediumBackground(&style->fg[i], &style->bg[i], MOTIF_FG_COLOR);

  /* Why do we need "mid" anyway?  None of my code touches it - and I
   * don't think any widgets do.  */
/*   style->mid[i].red = (style->light[i].red + style->dark[i].red) / 2; */
/*   style->mid[i].green = (style->light[i].green + style->dark[i].green) / 2; */
/*   style->mid[i].blue = (style->light[i].blue + style->dark[i].blue) / 2; */

  /* Allocate the new colours in the colormap for this style (or
   * something close to them). */
  if (! gdk_color_alloc (style->colormap, &style->light[i]))
    g_warning ("unable to allocate color: ( %d %d %d )",
               style->light[i].red, style->light[i].green,
               style->light[i].blue);
  if (! gdk_color_alloc (style->colormap, &style->dark[i]))
    g_warning ("unable to allocate color: ( %d %d %d )",
               style->dark[i].red, style->dark[i].green,
               style->dark[i].blue);
  if (! gdk_color_alloc (style->colormap, &style->mid[i]))
    g_warning ("unable to allocate color: ( %d %d %d )",
               style->mid[i].red, style->mid[i].green,
               style->mid[i].blue);
/* TODO: FOR SOME REASON, WHEN I SET THE FOREGROUND COLOR */
/* HERE, IT IS IMPOSSIBLE TO, LATER IN A PROGRAM, CREATE  */
/* A NEW GTKRCSTYLE AND THEN MODIFY THE FOREGROUND COLOR. */
/* FOR OTHER COLORS (LIKE DARK AND LIGHT, THIS IS POSSIBLE. */
/* (REALLY) */
/*   if (! gdk_color_alloc (style->colormap, &style->fg[i])) */
/*     g_warning ("unable to allocate color: ( %d %d %d )", */
/*                style->fg[i].red, style->fg[i].green, */
/*                style->fg[i].blue); */

  /* Get some new GCs with the new and improved colours. */
  gc_values.foreground = style->light[i];
  style->light_gc[i] = gtk_gc_get (style->depth, style->colormap,
                                   &gc_values, gc_values_mask);
  gc_values.foreground = style->dark[i];
  style->dark_gc[i] = gtk_gc_get (style->depth, style->colormap,
                                  &gc_values, gc_values_mask);
  gc_values.foreground = style->mid[i];
  style->mid_gc[i] = gtk_gc_get (style->depth, style->colormap,
                                 &gc_values, gc_values_mask);
/*   gc_values.foreground = style->fg[i]; */
/*   style->fg_gc[i] = gtk_gc_get (style->depth, style->colormap, */
/*                                  &gc_values, gc_values_mask); */
}
/* TODO: CALL MOTIF CALCULATIONS HERE */
static void
theme_realize_style(GtkStyle * style)
{
  calculate_colours(style, GTK_STATE_NORMAL);
  calculate_colours(style, GTK_STATE_ACTIVE);
  calculate_colours(style, GTK_STATE_PRELIGHT);
  calculate_colours(style, GTK_STATE_SELECTED);
  calculate_colours(style, GTK_STATE_INSENSITIVE);
}

static void
theme_unrealize_style(GtkStyle * style)
{
}

static void
theme_destroy_rc_style(GtkRcStyle * rc_style)
{
}

static void
theme_destroy_style(GtkStyle * style)
{
}

void
theme_init(GtkThemeEngine * engine)
{
  engine->parse_rc_style = theme_parse_rc_style;
  engine->merge_rc_style = theme_merge_rc_style;
  engine->rc_style_to_style = theme_rc_style_to_style;
  engine->duplicate_style = theme_duplicate_style;
  engine->realize_style = theme_realize_style;
  engine->unrealize_style = theme_unrealize_style;
  engine->destroy_rc_style = theme_destroy_rc_style;
  engine->destroy_style = theme_destroy_style;
  engine->set_background = NULL;

}

void
theme_exit(void)
{
}

/* The following function will be called by GTK+ when the module
 * is loaded and checks to see if we are compatible with the
 * version of GTK+ that loads us.
 */
G_MODULE_EXPORT const gchar* g_module_check_init (GModule *module);
const gchar*
g_module_check_init (GModule *module)
{
  return gtk_check_version (GTK_MAJOR_VERSION,
			    GTK_MINOR_VERSION,
			    GTK_MICRO_VERSION - GTK_INTERFACE_AGE);
}

